"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var AbortError_1 = require("./AbortError");
var defaultAbortControllerCtor_1 = require("./defaultAbortControllerCtor");
var AbortInCoroutines = /** @class */ (function () {
    function AbortInCoroutines(gen, _a) {
        var _this = this;
        var _b = _a === void 0 ? {} : _a, optSig = _b.signal, _c = _b.AbortController, AbortController = _c === void 0 ? defaultAbortControllerCtor_1.default : _c;
        this._ctrl = null;
        this._isAborted = false;
        if (!AbortController) {
            throw new TypeError('`AbortController` polyfill(or ponyfill) is needed.');
        }
        this._promise = new Promise(function (_resolve, _reject) {
            var offs = [];
            function on(sig, cb) {
                sig.addEventListener('abort', cb);
                offs.push(function () { return sig.removeEventListener('abort', cb); });
            }
            var cleanup = function () {
                offs.forEach(function (off) { return off(); });
                _this._ctrl = offs = null;
            };
            var resolve = function (val) { return (_resolve(val), cleanup()); };
            var reject = function (err) { return (_reject(err), cleanup()); };
            var abort = function (msg) {
                _this._isAborted = true;
                reject(new AbortError_1.default(msg));
            };
            if (optSig) {
                if (optSig.aborted) {
                    return abort('`options.signal` is already aborted.');
                }
                on(optSig, function () { return _this.abort(); });
            }
            var signal = (_this._ctrl = new AbortController()).signal;
            var iter = gen(signal);
            var pRunning = null;
            var done = resolve;
            on(signal, function () {
                if (pRunning && pRunning instanceof AbortInCoroutines)
                    pRunning.abort();
                pRunning = null;
                done = function (val) {
                    if (val === undefined) {
                        abort();
                    }
                    else {
                        resolve(val);
                    }
                };
                var res = iter.return(undefined);
                if (res.done) {
                    done(res.value);
                }
                else {
                    handle(res);
                }
            });
            function iterNext(arg) {
                var res;
                try {
                    res = iter.next(arg);
                }
                catch (err) {
                    return reject(err);
                }
                handle(res);
            }
            function iterThrow(arg) {
                var res;
                try {
                    res = iter.throw(arg);
                }
                catch (err) {
                    return reject(err);
                }
                handle(res);
            }
            function handle(res) {
                if (res.done) {
                    done(res.value);
                }
                else {
                    if (isThenable(res.value)) {
                        (pRunning = res.value).then(function (val) { return pRunning === res.value && iterNext(val); }, function (err) { return pRunning === res.value && iterThrow(err); });
                    }
                    else {
                        iterNext(res.value);
                    }
                }
            }
            iterNext();
        });
        this._promise.catch(noop); // prevent `unhandledrejection`
    }
    AbortInCoroutines.prototype.then = function (onfulfilled, onrejected) {
        return this._promise.then(onfulfilled, onrejected);
    };
    AbortInCoroutines.prototype.catch = function (onrejected) {
        return this._promise.catch(onrejected);
    };
    AbortInCoroutines.prototype.finally = function (onfinally) {
        return this._promise.finally(onfinally);
    };
    Object.defineProperty(AbortInCoroutines.prototype, "isAborted", {
        get: function () {
            return this._isAborted;
        },
        enumerable: true,
        configurable: true
    });
    AbortInCoroutines.prototype.abort = function () {
        var _a;
        (_a = this._ctrl) === null || _a === void 0 ? void 0 : _a.abort();
    };
    return AbortInCoroutines;
}());
exports.default = AbortInCoroutines;
function noop() { }
function isThenable(promise) {
    return promise && typeof promise.then === 'function';
}
